import pprint
import copy

from collections import defaultdict
from cStringIO import StringIO
from itertools import groupby

from numpy import array
from six.moves import xrange

from scrapely.extraction.pageobjects import AnnotationTag
from scrapely.extraction.regionextract import (
    RecordExtractor, BasicTypeExtractor, RepeatedDataExtractor,
    AdjacentVariantExtractor
)
from scrapely.extraction.similarity import (
    similar_region, longest_unique_subsequence, first_longest_subsequence
)
from scrapely.htmlpage import HtmlTagType
from scrapy.utils.spider import arg_to_iter

from .region_extractors import SlybotRecordExtractor, BaseExtractor
from .utils import container_id, group_tree, Region, element_from_page_index
from ..processors import MissingRequiredError, ItemProcessor


MAX_SEARCH_DISTANCE_MULTIPLIER = 3
MIN_TOKEN_LENGTH_BEFORE_TRUNCATE = 3
MIN_JUMP_DISTANCE = 0.7
MAX_RELATIVE_SEPARATOR_MULTIPLIER = 0.7


class BaseContainerExtractor(object):
    _extractor_classes = [
        RepeatedDataExtractor,
        AdjacentVariantExtractor,
        RepeatedDataExtractor,
        AdjacentVariantExtractor,
        RepeatedDataExtractor,
        RecordExtractor,
    ]

    def __init__(self, extractors, template):
        schema_name = self.annotation.metadata.get('schema_id')
        self.schema = template.descriptor(schema_name).copy()
        self.modifiers = template.modifiers
        extra_requires = getattr(self, 'extra_requires', [])
        self.extra_requires = extra_requires
        if hasattr(self.schema, '_required_attributes'):
            requires = list(extra_requires) + self.schema._required_attributes
            self.schema._required_attributes = requires

    @property
    def annotation(self):
        try:
            return self._annotation
        except AttributeError:
            return AnnotationTag(1, 1)

    @classmethod
    def apply(cls, template, extractors):
        # Group containers and get container info
        container_info = cls._get_container_data(extractors)
        containers, container_annos, non_container_annos = container_info
        # Build extraction tree
        extraction_tree = cls._build_extraction_tree(containers)
        tree = group_tree(extraction_tree, container_annos)

        # Build containerized extractors
        container_extractors = cls._build_containerized_extractors(
            containers, container_annos, template, tree)
        return non_container_annos + container_extractors

    def _build_extractors(self, extractors, containers, container_contents,
                          template):
        new_extractors = []
        annotation = None
        if isinstance(extractors, list):  # Bottom of extraction tree
            self.extra_requires = set()
            for ex in extractors:
                fields = ex.annotation.metadata.get('required_fields', [])
                self.extra_requires.update(fields)
            new_extractors.extend(
                SlybotRecordExtractor.apply(template, extractors))
        else:
            for container_name, container_data in (extractors or {}).items():
                annotation = self._find_annotation(template, container_name)
                if annotation.metadata.get('item_container'):
                    self._add_new_container(
                        annotation, new_extractors, container_data, template,
                        containers, container_contents)
        if not hasattr(self, '_annotation'):
            self._annotation = None
        if annotation:
            self._annotation = annotation
        elif new_extractors and self._annotation is None:
            self._annotation = new_extractors[0].annotation
        return new_extractors

    @staticmethod
    def _get_container_data(extractors):
        """
        Find information about all containers.

        Group all annotations by container.
        Group all annotations without a container.
        """
        # Get information about nested data
        containers = dict((x.annotation.metadata['id'], x)
                          for x in extractors
                          if x.annotation.metadata.get('item_container'))
        container_annos = defaultdict(list)
        non_container_annos = []
        for con_id, annotation in groupby(extractors, container_id):
            annotation = list(annotation)
            # Handle repeated container deletion but not parent container
            if (con_id and str(con_id).endswith('#parent') and
                    con_id not in containers):
                con_id = con_id.strip('#parent')
            if con_id and con_id in containers:
                container_annos[con_id].extend(annotation)
            else:
                non_container_annos.extend([
                    e for e in annotation
                    if e.annotation.metadata.get('id') not in containers
                ])
        return (containers, container_annos, non_container_annos)

    @staticmethod
    def _build_extraction_tree(containers):
        """
        Find the containers path for each sub container.
        """
        extraction_tree = []
        for container in containers.values():
            parent_id = container_id(container)
            path = [container.annotation.metadata['id']]
            parent = container
            if parent_id:
                while container_id(parent):
                    parent_id = container_id(parent)
                    if parent_id in path:
                        break  # Avoid cyclical tree
                    path.append(parent_id)
                    parent = containers[parent_id]
            path.reverse()
            extraction_tree.append(path)
        return extraction_tree

    @classmethod
    def _build_containerized_extractors(cls, containers, container_annos,
                                        template, tree):
        """
        Convert container annotation trees into container extractors.
        """
        container_extractors = []
        for container_name, container_data in tree.items():
            if container_name not in containers:
                continue  # Ignore missing containers
            container = container_annos[container_name]
            annotation = None
            for a in template.annotations:
                if a.metadata.get('id') == container_name:
                    annotation = a
                    break
            if container:
                cls._add_new_container(
                    annotation, container_extractors, container_data,
                    template, container_annos, container)
        return container_extractors

    @staticmethod
    def _add_new_container(annotation, extractors, container_data, template,
                           containers, container_contents):
        """
        Create a new container from the provided container information.
        """
        meta = annotation.metadata
        cls = ContainerExtractor
        if meta.get('repeated'):
            cls = RepeatedContainerExtractor
            if meta.get('field'):
                cls = RepeatedFieldsExtractor
        extractors.append(
            cls(container_data, template, annotation=annotation,
                containers=containers, container_contents=container_contents))

    def _find_annotation(self, template, annotation_id):
        """
        Look for an annotation with the given id in the given template
        """
        annotation_id = annotation_id.decode('utf-8')
        for annotation in template.annotations:
            aid = annotation.metadata.get('id', '').decode('utf-8')
            if aid == annotation_id:
                return annotation

    def _validate_and_adapt_item(self, item, htmlpage, region=None,
                                 surrounding_region=None):
        if isinstance(item, ItemProcessor):
            return item

        # Check if all required fields are available
        if isinstance(item, dict):
            for k in self.extra_requires:
                if k.startswith('_sticky'):  # Sticky no longer supported
                    item.pop(k, None)
                    continue
                if k not in item:
                    return {}
        # TODO: Pass item region
        processor = ItemProcessor(item, self, region, surrounding_region,
                                  htmlpage)

        # Check if item is valid
        dumped = processor.dump()
        if dumped:
            return processor
        else:
            return {}

    def __str__(self):
        stream = StringIO()
        pprint.pprint(self.extractors, stream)
        stream.seek(0)
        template_data = stream.read()
        if template_data:
            return "%s[\n%s\n]" % (self.__class__.__name__, template_data)
        return "%s[none]" % (self.__class__.__name__)


class ContainerExtractor(BaseContainerExtractor, BasicTypeExtractor):
    def __init__(self, extractors, template, containers=None,
                 container_contents=None, annotation=None,
                 required_fields=None):
        if containers is None:
            containers = {}
        if container_contents is None:
            container_contents = {}
        if annotation is not None:
            aid = annotation.metadata.get('id')
        else:
            aid = None
        self.template_tokens = template.page_tokens
        self.template_token_dict = template.token_dict
        self.extractors = self._build_extractors(
            extractors, containers, container_contents, template)
        record = [e for e in containers.get(aid, [])
                  if not e.annotation.metadata.get('item_container')]
        if record and not isinstance(extractors, list):
            self.extractors.extend(
                SlybotRecordExtractor.apply(template, record))
        self.content_validate = lambda x: x
        self.best_match = longest_unique_subsequence
        if annotation:
            self._annotation = annotation
        BaseContainerExtractor.__init__(self, extractors, template)
        if required_fields:
            self.extra_requires = set(required_fields) | self.extra_requires
        # A Container can only extract many items if it has at least one child
        # RepeatedContainerExtractor
        self.many = False
        if (any(isinstance(e, RepeatedContainerExtractor)
                for e in self.extractors)):
            self.many = True

    def extract(self, page, start_index=0, end_index=None,
                ignored_regions=None, **kwargs):
        """
        Find a region surrounding repeated data and run extractors on the data
        in that region.
        """
        start_index = max(0, start_index - 1)
        max_end_index = len(page.token_page_indexes)
        if end_index is None:
            end_index = max_end_index
        else:
            end_index = min(max_end_index, end_index + 1)
        region = Region(*similar_region(
            page.page_tokens, self.template_tokens, self.annotation,
            start_index, end_index, self.best_match, **kwargs))
        if region.score < 1:
            return []
        surrounding = element_from_page_index(page, start_index)
        items = self._extract_items_from_region(
            region, page, ignored_regions, surrounding, **kwargs)
        tag = element_from_page_index(page, region.start_index)
        items = [self._validate_and_adapt_item(i, page, tag) for i in items]
        if self.many:
            return items
        return self._merge_items(items)

    def _extract_items_from_region(self, region, page, ignored_regions,
                                   surrounding, **kwargs):
        items = []
        for extractor in self.extractors:
            try:
                try:
                    item = extractor.extract(
                        page, region.start_index, region.end_index,
                        ignored_regions, **kwargs
                    )
                except TypeError:
                    ex = SlybotRecordExtractor(
                        extractor.extractors, extractor.template_tokens
                    )
                    item = ex.extract(
                        page, region.start_index, region.end_index,
                        ignored_regions, **kwargs
                    )
            except MissingRequiredError:
                return []
            if (isinstance(extractor, BaseContainerExtractor) and
                    isinstance(item, list)):
                items.extend(item)
            else:
                if not isinstance(item, ItemProcessor):
                    item = ItemProcessor(item, self, [region], surrounding,
                                         page)
                items.append(item)
        return items

    def _merge_items(self, items):
        if not items:
            return []
        item = items[0]
        for other_item in items[1:]:
            item.merge(other_item)
        return item


class RepeatedContainerExtractor(BaseContainerExtractor, RecordExtractor):
    def __init__(self, extractors, template, containers=None,
                 container_contents=None, annotation=None, schemas=None):
        if containers is None:
            containers = {}
        if container_contents is None:
            container_contents = {}
        if schemas is None:
            schemas = {}
        if annotation is not None:
            aid = annotation.metadata.get('id')
        else:
            aid = None
        self.template_tokens = template.page_tokens
        self.template_token_dict = template.token_dict
        self.prefix, self.suffix = self._find_prefix_suffix(
            extractors, container_contents, containers, template)
        self.extractors = self._build_extractors(
            extractors, containers, container_contents, template)
        record = [e for e in containers.get(aid, [])
                  if not e.annotation.metadata.get('item_container')]
        if record and not isinstance(extractors, list):
            self.extractors.extend(
                SlybotRecordExtractor.apply(template, record))
        self.best_match = first_longest_subsequence
        BaseContainerExtractor.__init__(self, extractors, template)

    def extract(self, page, start_index=0, end_index=None,
                ignored_regions=None, **kwargs):
        """
        Find regions bounded by the prefix and suffix and repeatedly
        extract them.
        """
        prefixlen = len(self.prefix)
        suffixlen = len(self.suffix)
        index = max(0, start_index - prefixlen)
        if end_index is None:
            end_index = len(self.template_tokens)
        max_index = min(len(page.page_tokens) - suffixlen,
                        end_index + len(self.suffix))
        max_start_index = max_index - prefixlen
        extracted = []
        surrounding_tag = element_from_page_index(page, start_index)
        while index <= max_start_index:
            prefix_end = index + prefixlen
            if (page.page_tokens[index:prefix_end] == self.prefix).all():
                for peek in xrange(prefix_end + self.min_jump, max_index + 1):
                    next_prefix = page.page_tokens[peek:peek + prefixlen]
                    next_suffix = page.page_tokens[peek:peek + suffixlen]
                    matches_next_prefix = (next_prefix == self.prefix).all()
                    matches_next_suffix = (next_suffix == self.suffix).all()
                    if not (matches_next_suffix or matches_next_prefix or
                            peek + suffixlen >= max_index):
                        continue
                    if matches_next_prefix:
                        peek -= suffixlen + 1
                    try:
                        items = []
                        _index = index
                        for extractor in self.extractors:
                            items += extractor.extract(
                                page, index, peek + self.offset,
                                ignored_regions,
                                suffix_max_length=suffixlen)
                            _index = max(peek, index) - 1
                    except MissingRequiredError:
                        pass
                    else:
                        tag = element_from_page_index(page, index)
                        processed = self._process_items(items, page, tag,
                                                        surrounding_tag)
                        extracted.extend(processed)
                    index = _index
                    break
            index += 1
        result = []
        for i, item in enumerate(extracted, 1):
            if not item:
                continue
            try:
                item[u'_index'] = i
            except TypeError:
                pass
            result.append(item)
        if self.parent_annotation.metadata.get('field'):
            return [(self.parent_annotation.metadata['field'], result)]
        return result

    def _process_items(self, items, page, region, surrounding_region):
        if not items:
            return []
        items = self._validate_and_adapt_item(items, page, region,
                                              surrounding_region)
        return arg_to_iter(items)

    def _find_prefix_suffix(self, extractors, container_contents, containers,
                            template):
        """
        Find the prefix and suffix for this repeating extractor.
        """
        htt = HtmlTagType
        all_tags = (htt.CLOSE_TAG, htt.UNPAIRED_TAG, htt.OPEN_TAG)
        open_tags = (htt.OPEN_TAG, htt.UNPAIRED_TAG)
        parent, child = self._find_siblings(template, containers,
                                            container_contents)
        self.parent_annotation = parent
        parent_sindex = 0 if not parent else parent.start_index
        tokens = template.page_tokens[parent_sindex:child.start_index + 1]
        prefix = self._find_tokens(tokens, open_tags, template)
        prefix.reverse()
        tokens = template.page_tokens[child.start_index + 1:
                                      child.end_index + 1]
        suffix = self._find_tokens(tokens, all_tags, template)
        prefix = self._trim_prefix(prefix, suffix, template)
        suffix.reverse()
        suffix = self._trim_prefix(suffix, prefix, template, 3)
        tokens = template.page_tokens[child.start_index + 1:
                                      child.end_index][::-1]
        max_separator = child.metadata.get('max_separator', -1)
        if max_separator == -1:
            max_separator = int(
                len(tokens) * MAX_RELATIVE_SEPARATOR_MULTIPLIER)
        tokens = self._find_tokens(tokens, open_tags, template)
        prefix = self._trim_prefix(
            prefix + tokens, suffix, template, 3, True)
        tokens = template.page_tokens[child.end_index + 1:
                                      child.end_index + max_separator][::-1]
        tokens = self._find_tokens(tokens, open_tags, template, prefix[0])
        self.offset = 1 if not tokens else 0
        suffix = self._trim_prefix(suffix + tokens, prefix, template, 3, True)
        # Heuristic to reduce chance of false positives
        self.min_jump = child.metadata.get('min_jump', -1)
        if self.min_jump == -1:
            self.min_jump = int((child.end_index - child.start_index -
                                 len(suffix)) * MIN_JUMP_DISTANCE)
        return (array(prefix), array(suffix))

    def _find_siblings(self, template, containers, container_contents):
        child_id = container_contents[0].annotation.metadata['container_id']
        child = self._annotation = containers[child_id][0].annotation
        parent_id = child.metadata.get('container_id')
        parent = self._find_annotation(template, parent_id)
        siblings = child.metadata.get('siblings', 0)
        end = child.end_index
        if siblings > 0:
            end = self._find_siblings_end(template, child.end_index + 1,
                                          parent.end_index, siblings)
        if end is not None:
            new_child = AnnotationTag(child.start_index, end,
                                      child.surrounds_attribute,
                                      child.annotation_text,
                                      child.tag_attributes)
            new_child.metadata = child.metadata
            self._annotation = new_child
            return (parent, new_child)
        return parent, child

    def _find_siblings_end(self, template, start_index, max_index, siblings):
        if start_index >= max_index:
            return None
        if siblings == 0:
            return start_index
        htt = HtmlTagType
        first_token = template.page_tokens[start_index]
        if template.token_dict.token_type(first_token) == htt.UNPAIRED_TAG:
            return self._find_siblings_end(template, start_index + 1,
                                           max_index, siblings - 1)
        tokens = template.page_tokens[start_index + 1:max_index]
        tag_stack = [first_token]
        for idx, token in enumerate(tokens, 1):
            token_type = template.token_dict.token_type(token)
            if token_type == htt.OPEN_TAG:
                tag_stack.append(token)
            elif token_type == htt.CLOSE_TAG:
                tag_stack.pop(-1)
            if not tag_stack:
                return self._find_siblings_end(template, start_index + idx + 1,
                                               max_index, siblings - 1)

    def _trim_prefix(self, prefix, suffix, template, min_prefix_len=1,
                     remove_from_end=False):
        """
        If the annotation is on the first element it is possible that there are
        too many tokens in the prefix to find sibling regions. This looks ahead
        for sibling regions and reduces the number of tokens until it can find
        some.
        """
        if len(prefix) <= min_prefix_len:
            return prefix
        page_tokens = template.page_tokens
        new_prefix = copy.copy(prefix)
        suffixlen = len(suffix)
        end_index = self.annotation.end_index
        start_index = self.annotation.start_index
        annotation_length = start_index - end_index
        max_start_index = min(
            start_index + MAX_SEARCH_DISTANCE_MULTIPLIER * annotation_length,
            len(page_tokens))
        while len(new_prefix) > min_prefix_len:
            index = end_index
            while index < max_start_index:
                prefixlen = len(new_prefix)
                prefix_end = index + prefixlen
                if (page_tokens[index:prefix_end] == new_prefix).all():
                    for peek in xrange(prefix_end, max_start_index + 1):
                        suffix_tokens = page_tokens[peek:peek + suffixlen]
                        if (suffix_tokens == suffix).all():
                            return new_prefix
                index += 1
            if remove_from_end:
                new_prefix = new_prefix[:-1]
            else:
                new_prefix = new_prefix[1:]

        return new_prefix

    @staticmethod
    def _find_tokens(tokens, token_types, template, upto=None):
        """
        Find a consecutive list tokens marching the supplied token types.
        Possibly remove the final token as this may reduce the number of
        regions that can be found.
        """
        result_tokens = []
        token_type = template.token_dict.token_type
        for token in reversed(tokens):
            if token == upto:
                break
            if token_type(token) in token_types:
                result_tokens.append(token)
            else:
                break
        if len(result_tokens) > MIN_TOKEN_LENGTH_BEFORE_TRUNCATE:
            result_tokens = result_tokens[:-1]
        return result_tokens


class RepeatedFieldsExtractor(RepeatedContainerExtractor):
    def __init__(self, extractors, template, containers=None,
                 container_contents=None, annotation=None, schemas=None):
        extractors = [BaseExtractor(annotation, schemas)]
        self._annotation = annotation
        RepeatedContainerExtractor.__init__(
            self, extractors, template, containers, container_contents,
            schemas)
        self.extractors = [RepeatedDataExtractor(
            [self.prefix[-2:]], [self.suffix[:1]],
            [BaseExtractor(annotation, self.schema)])]

    def extract(self, page, start_index=0, end_index=None,
                ignored_regions=None, **kwargs):
        data = self.extractors[0].extract(
            page, start_index, end_index, ignored_regions, **kwargs)
        if not data:
            return []
        values = [v for _, v in data]
        return [(data[0][0], values)]

    def _validate_and_adapt_item(self, item, htmlpage=None, region=None,
                                 surrounding_region=None):
        return item

    def _find_siblings(self, template, containers, container_contents):
        container_id = self.annotation.metadata['container_id']
        parent = self._find_annotation(template, container_id)
        return parent, self.annotation



from unittest import TestCase

from slybot.starturls import FragmentGenerator


class FragmentGeneratorTest(TestCase):
    def test_generated_url_list(self):
        github_start_urls = [
            'https://github.com/scrapinghub',
            'https://github.com/scrapy',
            'https://github.com/scrapy-plugins'
        ]
        url_spec = {
            'fragments': [
                {'type': 'fixed', 'value': 'https://github.com/'},
                {'type': 'list', 'value': 'scrapinghub scrapy scrapy-plugins'},
            ]
        }
        generator = FragmentGenerator()

        self.assertEqual(list(generator(url_spec)), github_start_urls)

    def test_generated_url_range(self):
        github_start_urls = [
            'https://github.com/0',
            'https://github.com/1',
            'https://github.com/2'
        ]
        url_spec = {
            'fragments': [
                {'type': 'fixed', 'value': 'https://github.com/'},
                {'type': 'range', 'value': '0-2'},
            ]
        }
        generator = FragmentGenerator()

        self.assertEqual(list(generator(url_spec)), github_start_urls)

    def test_mixed_fragments(self):
        donedeal_start_urls = [
            'https://www.donedeal.ie/cars-for-sale/i/1',
            'https://www.donedeal.ie/cars-for-sale/i/2',
            'https://www.donedeal.ie/houses-for-sale/i/1',
            'https://www.donedeal.ie/houses-for-sale/i/2',
            'https://www.donedeal.ie/pets-for-sale/i/1',
            'https://www.donedeal.ie/pets-for-sale/i/2',
            'https://www.donedeal.ie/kitchens-for-sale/i/1',
            'https://www.donedeal.ie/kitchens-for-sale/i/2'
        ]
        url_spec = {
            'fragments': [
                {'type': 'fixed', 'value': 'https://www.donedeal.ie/'},
                {
                    'type': 'list',
                    'value': ('cars-for-sale houses-for-sale '
                              'pets-for-sale kitchens-for-sale')
                },
                {'type': 'fixed', 'value': '/i/'},
                {'type': 'range', 'value': '1-2'},
            ]
        }
        generator = FragmentGenerator()

        self.assertEqual(list(generator(url_spec)), donedeal_start_urls)

    def test_generated_letters(self):
        github_start_urls = [
            'https://github.com/scrapinghub/a',
            'https://github.com/scrapinghub/b',
            'https://github.com/scrapinghub/c',
            'https://github.com/scrapinghub/d',
        ]
        url_spec = {
            'fragments': [
                {'type': 'fixed', 'value': 'https://github.com/scrapinghub/'},
                {'type': 'range', 'value': 'a-d'},
            ]
        }
        generator = FragmentGenerator()

        self.assertEqual(list(generator(url_spec)), github_start_urls)



import json
import re

LUA_SOURCE = """
function main(splash)
    assert(splash:go(splash.args.url))
    splash:runjs(splash.args.js_source)
    splash:wait_for_resume(splash.args.slybot_actions_source)
    splash:set_result_content_type("text/html")
    return splash.html()
end
"""

JS_SOURCE = """
function main(splash) {
    var events = (%s);
    try{
        __slybot__performEvents(events, function(){
            splash.resume();
        });
    }catch(e){
        splash.error(e);
    }
}
"""

def filter_for_url(url):
    def _filter(page_action):
        accept = page_action.get('accept')
        reject = page_action.get('reject')
        if reject and re.search(reject, url):
            return False
        if accept and not re.search(accept, url):
            return False
        return True
    return _filter

class PageActionsMiddleware(object):
    def process_request(self, request, spider):
        splash_options = request.meta.get('splash', None)
        if not splash_options: # Already processed or JS disabled
            return
        splash_args = splash_options.get('args', {})
        events = spider.page_actions
        url = splash_args['url']
        events = filter(filter_for_url(url), events)
        if len(events):
            splash_options['endpoint'] = 'execute'
            splash_args.update({
                "lua_source": LUA_SOURCE,
                "slybot_actions_source": (JS_SOURCE % json.dumps(events)),
            })

__all__ = ['PageActionsMiddleware']




import re

from scrapely.extractors import htmlregion
from scrapely.htmlpage import HtmlPageRegion

from slybot.fieldtypes import FieldTypeManager
from slybot.item import SlybotFieldDescriptor, SlybotItemDescriptor


def create_regex_extractor(pattern):
    r"""Create extractor from a regular expression.

    Only groups from match are extracted and concatenated, so it
    is required to define at least one group. Ex:
    >>> extractor = create_regex_extractor("(\d+).*(\.\d+)")
    >>> extractor(u"The price of this product is <div>45</div> </div class='small'>.50</div> pounds")
    u'45.50'
    """
    ereg = re.compile(pattern, re.S)

    def _extractor(txt, htmlpage=None):
        if txt is None:
            return
        m = ereg.search(txt)
        if m:
            return htmlregion(u"".join([g for g in m.groups() or m.group()
                                        if g]))

    _extractor.__name__ = "Regex: %s" % pattern.encode("utf-8")
    return _extractor


def create_type_extractor(_type):
    types = FieldTypeManager()
    extractor = types.type_processor_class(_type)()

    def _extractor(txt, htmlpage=None):
        if txt is None:
            return
        page = getattr(htmlpage, 'htmlpage', htmlpage)
        if not hasattr(txt, 'text_content'):
            txt = HtmlPageRegion(page, txt)
        data = extractor.extract(txt)
        if data:
            return extractor.adapt(data, page)
    _extractor.__name__ = ("Type Extractor: %s" % _type).encode('utf-8')
    return _extractor


class PipelineExtractor:
    def __init__(self, *extractors):
        self.extractors = extractors

    def __call__(self, value):
        for extractor in self.extractors:
            value = extractor(value) if value else value
        return value

    @property
    def __name__(self):
        return repr(self.extractors)


def apply_extractors(descriptor, template_extractors, extractors):
    type_processor_class = FieldTypeManager().type_processor_class
    if isinstance(template_extractors, dict):
        template_extractors = template_extractors.items()
    attribute_map = descriptor.attribute_map
    for field_name, field_extractors in template_extractors:
        equeue = []
        for eid in field_extractors:
            e_doc = extractors.get(eid, {})
            if "regular_expression" in e_doc:
                equeue.append(
                    create_regex_extractor(e_doc["regular_expression"]))
            elif "type_extractor" in e_doc:  # overrides default one
                try:
                    display_name = attribute_map[field_name].description
                except KeyError:
                    display_name = field_name
                field_type = type_processor_class(e_doc["type_extractor"])()
                attribute_map[field_name] = SlybotFieldDescriptor(
                    field_name, display_name, field_type)
        if field_name not in attribute_map:
            # if not defined type extractor, use text type by default, as it is
            # by far the most commonly used
            attribute_map[field_name] = SlybotFieldDescriptor(
                field_name, field_name,
                type_processor_class("text")())

        if equeue:
            equeue.insert(0, attribute_map[field_name].extractor)
            attribute_map[field_name].extractor = PipelineExtractor(*equeue)


def add_extractors_to_descriptors(descriptors, extractors):
    new_extractors = {}
    for _id, data in extractors.items():
        if "regular_expression" in data:
            extractor = create_regex_extractor(data['regular_expression'])
        else:
            extractor = create_type_extractor(data['type_extractor'])
        new_extractors[_id] = extractor
    for descriptor in descriptors.values():
        if isinstance(descriptor, SlybotItemDescriptor):
            descriptor.extractors = new_extractors



"""
Price field types
"""
from scrapely import extractors

class PriceTypeProcessor(object):
    """Extracts price from text"""
    name = "price"
    description = "extracts a price decimal number in the text passed"

    def extract(self, htmlregion):
        return extractors.contains_any_numbers(htmlregion.text_content)

    def adapt(self, text, htmlpage=None):
        return extractors.extract_price(text)




from unittest import TestCase

from slybot.starturls import FragmentGenerator, IdentityGenerator, StartUrlCollection, UrlGenerator


class StartUrlCollectionTest(TestCase):
    def setUp(self):
        self.generators = {
            'start_urls': IdentityGenerator(),
            'generated_urls': UrlGenerator(),
            'url': IdentityGenerator(),
            'generated': FragmentGenerator(),
        }

    def test_mixed_start_urls_generation(self):
        start_urls = [
            'http://google.com',
            {"type": "url", "url": "http://domain.com"},
            {
                'type': 'generated',
                'url': 'https://github.com/[0-2]',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://github.com/'},
                    {'type': 'range', 'value': '0-2'},
                ]
            }
        ]
        generated_start_urls = [
            'http://google.com',
            'http://domain.com',
            'https://github.com/0',
            'https://github.com/1',
            'https://github.com/2',
        ]

        generated = StartUrlCollection(start_urls, self.generators)
        self.assertEqual(list(generated), generated_start_urls)

    def test_generated_type(self):
        generated_start_urls = [
            'https://github.com/scrapinghub',
            'https://github.com/scrapy',
            'https://github.com/scrapy-plugins',
        ]
        start_urls = [
            {
                "template": "https://github.com/{}",
                "paths": [{
                    "type": "options",
                    "values": ["scrapinghub", "scrapy", "scrapy-plugins"],
                }],
                "params": [],
                "params_template": {}
            },
        ]
        generated = StartUrlCollection(start_urls, self.generators)

        self.assertEqual(list(generated), generated_start_urls)

    def test_unique_legacy_urls(self):
        start_urls = [
            'http://google.com',
            'http://github.com',
            'http://github.com',
            'http://scrapinghub.com',
            'http://scrapinghub.com',
        ]
        unique_urls = [
            'http://google.com',
            'http://github.com',
            'http://scrapinghub.com',
        ]

        self.assertEqual(StartUrlCollection(start_urls).uniq(), unique_urls)

    def test_unique_list_start_urls(self):
        start_urls = [
            {"type": "url", "url": "http://domain.com"},
            {
                'type': 'generated',
                'url': 'https://github.com/[...]',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://github.com/'},
                    {'type': 'list', 'value': 'scrapely portia'},
                ]
            },
            {
                'type': 'generated',
                'url': 'https://github.com/[...]',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://github.com/'},
                    {'type': 'list', 'value': 'scrapely scrapinghub portia'},
                ]
            },
        ]

        self.assertEqual(StartUrlCollection(start_urls).uniq(), start_urls)

    def test_allowed_domains_with_many_fragments(self):
        start_urls = [
            {
                'type': 'generated',
                'url': 'https://github.com/[...]',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://github.com'},
                    {'type': 'list', 'value': '/a /b /c'},
                    {'type': 'range', 'value': '1-10000000'},
                ]
            },
        ]
        allowed_domains = [
            'https://github.com/a',
            'https://github.com/b',
            'https://github.com/c',
        ]
        collection_domains = StartUrlCollection(start_urls, self.generators).allowed_domains
        self.assertEqual(set(collection_domains), set(allowed_domains))

    def test_allowed_domains_with_mixed_urls(self):
        start_urls = [
            {
                'type': 'generated',
                'url': 'https://scrapinghub.com/[...]',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://scrapinghub.com/'},
                    {'type': 'range', 'value': '1-10000000'},
                ]
            },
            {
                'type': 'generated',
                'url': 'https://github[1-3].com/[...]',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://github'},
                    {'type': 'range', 'value': '1-3'},
                    {'type': 'fixed', 'value': '.com/'},
                    {'type': 'range', 'value': '1-10000000'},
                ]
            },
            {"type": "url", "url": "http://domain.com"},
            'http://google.com',
        ]
        allowed_domains = [
            'https://scrapinghub.com/',
            'https://github1.com/',
            'https://github2.com/',
            'https://github3.com/',
            'http://domain.com',
            'http://google.com',
        ]
        collection_domains = StartUrlCollection(start_urls, self.generators).allowed_domains
        self.assertEqual(set(collection_domains), set(allowed_domains))

    def test_empty_allowed_domains(self):
        start_urls = [
            {
                'type': 'generated',
                'url': 'https://',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://'},
                ]
            },
        ]
        collection_domains = StartUrlCollection(start_urls, self.generators).allowed_domains
        self.assertEqual(collection_domains, [])

    def test_multiple_empty_allowed_domains(self):
        start_urls = [
            {
                'type': 'generated',
                'url': 'https://',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://'},
                    {'type': 'fixed', 'value': 'scrapy'},
                ]
            },
        ]
        collection_domains = StartUrlCollection(start_urls, self.generators).allowed_domains
        self.assertEqual(collection_domains, [])

    def test_normalize_string_url(self):
        legacy = ['https://github.com/scrapinghub']
        normalized = [{
            'url': 'https://github.com/scrapinghub',
            'type': 'url',
        }]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(legacy[0], normalized[0]['url'])
        self.assertEqual(list(collection.normalize()), normalized)

    def test_normalize_start_url(self):
        start_urls = [{
            'url': 'https://github.com/scrapinghub',
            'type': 'fixed',
        }]
        collection = StartUrlCollection(start_urls, self.generators)

        self.assertEqual(list(collection.normalize()), start_urls)

    def test_normalize_generated_options(self):
        legacy = [{
            "template": "https://github.com/{}",
            "paths": [{
                "type": "options",
                "values": ["scrapinghub", "scrapy", "scrapy-plugins"],
            }],
            "params": [],
            "params_template": {}
        }]
        normalized = [{
            'url': 'https://github.com/[...]',
            'type': 'generated',
            'fragments': [
                {'type': 'fixed', 'value': 'https://github.com/'},
                {
                    'type': 'list',
                    'value': 'scrapinghub scrapy scrapy-plugins',
                },
            ]
        }]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(generator_set(UrlGenerator, legacy[0]),
                         generator_set(FragmentGenerator, normalized[0]))
        self.assertEqual(list(collection.normalize()), normalized)

    def test_normalize_generated_default(self):
        legacy = [{
            "template": "https://github.com/{}/fixed",
            "paths": [{
                "type": "default",
                "values": ["scrapinghub", "scrapy", "scrapy-plugins"],
            }],
            "params": [],
            "params_template": {}
        }]
        normalized = [{
            'url': 'https://github.com/scrapinghub/fixed',
            'type': 'generated',
            'fragments': [
                {'type': 'fixed', 'value': 'https://github.com/'},
                {
                    'type': 'fixed',
                    'value': 'scrapinghub',
                },
                {'type': 'fixed', 'value': '/fixed'},
            ]
        }]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(generator_set(UrlGenerator, legacy[0]),
                         generator_set(FragmentGenerator, normalized[0]))
        self.assertEqual(list(collection.normalize()), normalized)

    def test_normalize_generated_dates(self):
        legacy = [{
            "template": "http://www.commitstrip.com/{}/{}/{}",
            "paths": [{
                "type": "default",
                "values": ["en"]
            }, {
                "type": "date",
                "values": ["%Y"],
            }, {
                "type": "date",
                "values": ["%m"]
            }],
            "params": [],
            "params_template": {}
        }]
        normalized = [{
            'url': 'http://www.commitstrip.com/en/[...]/[...]',
            'type': 'generated',
            'fragments': [
                {'type': 'fixed', 'value': 'http://www.commitstrip.com/'},
                {'type': 'fixed', 'value': 'en'},
                {'type': 'fixed', 'value': '/'},
                {'type': 'date', 'value': '%Y'},
                {'type': 'fixed', 'value': '/'},
                {'type': 'date', 'value': '%m'},
            ]
        }]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(generator_set(UrlGenerator, legacy[0]),
                         generator_set(FragmentGenerator, normalized[0]))
        self.assertEqual(list(collection.normalize()), normalized)

    def test_normalized_generated_range(self):
        legacy = [{
            "template": "https://www.donedeal.ie/{}/{}/{}",
            "paths": [{
                "type": "default",
                "values": ["cars-for-sale"]
            }, {
                "type": "options",
                "values": ["i"],
            }, {
                "type": "range",
                "values": [10, 20]
            }],
            "params": [],
            "params_template": {}
        }]
        normalized = [{
            'url': 'https://www.donedeal.ie/cars-for-sale/[...]/10-19',
            'type': 'generated',
            'fragments': [
                {'type': 'fixed', 'value': 'https://www.donedeal.ie/'},
                {'type': 'fixed', 'value': 'cars-for-sale'},
                {'type': 'fixed', 'value': '/'},
                {'type': 'list', 'value': 'i'},
                {'type': 'fixed', 'value': '/'},
                {'type': 'range', 'value': '10-19'},
            ]
        }]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(generator_set(UrlGenerator, legacy[0]),
                         generator_set(FragmentGenerator, normalized[0]))
        self.assertEqual(list(collection.normalize()), normalized)

    def test_normalized_generated_params_range(self):
        legacy = [{
            "template": "http://www.smbc-comics.com/{}",
            "paths": [{
                "type": "default",
                "values": ["index.php"]
            }],
            "params": [{
                "name": "p",
                "type": "range",
                "values": [20, 31]
            }, {
                "name": "q",
                "type": "options",
                "values": ['comic']
            }],
            "params_template": {}
        }]
        normalized = [{
            'url': 'http://www.smbc-comics.com/index.php?p=20-30&q=[...]',
            'type': 'generated',
            'fragments': [
                {'type': 'fixed', 'value': 'http://www.smbc-comics.com/'},
                {'type': 'fixed', 'value': 'index.php'},
                {'type': 'fixed', 'value': '?p='},
                {'type': 'range', 'value': '20-30'},
                {'type': 'fixed', 'value': '&q='},
                {'type': 'list', 'value': 'comic'},
            ]
        }]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(generator_set(UrlGenerator, legacy[0]),
                         generator_set(FragmentGenerator, normalized[0]))
        self.assertEqual(list(collection.normalize()), normalized)

    def test_normalized_generated_template_params(self):
        legacy = [{
            "template": "https://encrypted.google.com/search",
            "paths": [],
            "params": [{
                "name": "q",
                "type": "options",
                "values": ["nosetests", "tox"]
            }, {
                "name": "location",
                "type": "options",
                "values": ["dublin", "cork"]
            }],
            "params_template": [
                ("hl", "en"),
                ("q", "python unittest")
            ]
        }]
        normalized = [{
            'url': 'https://encrypted.google.com/search?hl=en&q=[...]&location=[...]',
            'type': 'generated',
            'fragments': [
                {'type': 'fixed', 'value': 'https://encrypted.google.com/search'},
                {'type': 'fixed', 'value': '?hl='},
                {'type': 'fixed', 'value': 'en'},
                {'type': 'fixed', 'value': '&q='},
                {'type': 'list', 'value': 'nosetests tox'},
                {'type': 'fixed', 'value': '&location='},
                {'type': 'list', 'value': 'dublin cork'},
            ]
        }]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(generator_set(UrlGenerator, legacy[0]),
                         generator_set(FragmentGenerator, normalized[0]))
        self.assertEqual(list(collection.normalize()), normalized)

    def test_normalized_mixed(self):
        legacy = [
            {
                "template": "http://www.smbc-comics.com/{}",
                "paths": [{
                    "type": "default",
                    "values": ["index.php"]
                }],
                "params": [{
                    "name": "p",
                    "type": "range",
                    "values": [20, 31]
                }, {
                    "name": "q",
                    "type": "options",
                    "values": ['comic']
                }],
                "params_template": {}
            },
            'http://github.com/scrapinghub.com',
            {
                'url': 'https://github.com/[...]',
                'type': 'generated',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://github.com/'},
                    {
                        'type': 'list',
                        'value': 'scrapinghub scrapy scrapy-plugins',
                    },
                ]
            }
        ]
        normalized = [
            {
                'url': 'http://www.smbc-comics.com/index.php?p=20-30&q=[...]',
                'type': 'generated',
                'fragments': [
                    {'type': 'fixed', 'value': 'http://www.smbc-comics.com/'},
                    {'type': 'fixed', 'value': 'index.php'},
                    {'type': 'fixed', 'value': '?p='},
                    {'type': 'range', 'value': '20-30'},
                    {'type': 'fixed', 'value': '&q='},
                    {'type': 'list', 'value': 'comic'},
                ]
            },
            {'url': 'http://github.com/scrapinghub.com', 'type': 'url'},
            {
                'url': 'https://github.com/[...]',
                'type': 'generated',
                'fragments': [
                    {'type': 'fixed', 'value': 'https://github.com/'},
                    {
                        'type': 'list',
                        'value': 'scrapinghub scrapy scrapy-plugins',
                    },
                ]
            },
        ]
        collection = StartUrlCollection(legacy, self.generators)

        self.assertEqual(list(collection.normalize()), normalized)

def generator_set(generator, start_urls):
    return set(list(generator()(start_urls)))



from unittest import TestCase
from scrapely.htmlpage import HtmlPage

from slybot.fieldtypes import UrlFieldTypeProcessor, ImagesFieldTypeProcessor

class FieldTypesUrlEncoding(TestCase):
    def test_not_standard_chars_in_url(self):
        body = u'<html><meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" /></html>'
        url = u'fotos/produtos/Mam\xe3e noel.jpg'
        htmlpage = HtmlPage(url=u"http://www.example.com/", body=body, encoding='cp1252')
        processor = UrlFieldTypeProcessor()
        self.assertEqual(processor.adapt(url, htmlpage), u'http://www.example.com/fotos/produtos/Mam%C3%A3e%20noel.jpg')

    def test_uri_stripped_of_whitespace_and_quote_characters_correctly(self):
        urls = [u' image.jpg ', u"    '/data.jpg'", u'\n\t"file.jpg"\n\t\t']
        results = ['http://www.example.com/images/image.jpg',
                   'http://www.example.com/data.jpg',
                   'http://www.example.com/images/file.jpg']
        htmlpage = HtmlPage(url=u"http://www.example.com/images/",
                            body=u'<html><body></body></html>',
                            encoding='utf-8')
        url_p = UrlFieldTypeProcessor()
        img_p = ImagesFieldTypeProcessor()
        for text, url in zip(urls, results):
            self.assertEqual(img_p.adapt(img_p.extract(text), htmlpage), url)
            self.assertEqual(url_p.adapt(url_p.extract(text), htmlpage), url)

    def test_uri_with_illegal_html_entities(self):
        urls = [u'&#9;&#10 image.jpg ', u"    '/&#11;&#0;data.jpg'",
                u'&#15;\n\t"&#14;file.jpg"\n\t\t']
        results = ['http://www.example.com/images/image.jpg',
                   'http://www.example.com/data.jpg',
                   'http://www.example.com/images/file.jpg']
        htmlpage = HtmlPage(url=u"http://www.example.com/images/",
                            body=u'<html><body></body></html>',
                            encoding='utf-8')
        url_p = UrlFieldTypeProcessor()
        img_p = ImagesFieldTypeProcessor()
        for text, url in zip(urls, results):
            self.assertEqual(img_p.adapt(img_p.extract(text), htmlpage), url)
            self.assertEqual(url_p.adapt(url_p.extract(text), htmlpage), url)

    def test_blank_image_url(self):
        assert ImagesFieldTypeProcessor().extract(' ') == ''
